package com.goldencode.p2j.ui;

import java.io.*;
import java.util.*;

import org.apache.commons.logging.*;
import org.dom4j.*;
import org.dom4j.Element;
import org.dom4j.io.SAXReader;
import org.dom4j.tree.DefaultElement;
import org.xml.sax.*;


public class ColorMapLoader
{
   private static final Log LOG = LogFactory.getLog(ColorMapLoader.class);
   
   private static final String COLOR_MAP_XML_RESOURCE = "/color-map.xml";
   private static final String COLOR_MAP_DTD_RESOURCE = "/color-map.dtd";
   private static final String COLOR_MAP_DTD_ID = "color-map.dtd";
   
   private static final String E_NAME_TYPES = "types";
   private static final String E_NAME_NAME = "name";
   private static final String E_NAME_FOREGROUND = "foreground";
   private static final String E_NAME_BACKGROUND = "background";
   private static final String E_NAME_ATTRIBUTES = "attributes";

   
   public Map<String, Map<String, Color>> loadAllMappings()
   {
      Document document = parseDocument();
      if (document == null)
      {
         return null;
      }

      Map<String, Map<String, Color>> output = new HashMap<String, Map<String, Color>>();

      for (Iterator<?> i = document.getRootElement().elementIterator(); i.hasNext();)
      {
         Object element = i.next();

         if (!(element instanceof Element))
         {
            continue;
         }

         Element ansiTerminal = (Element) element;
         String types = ansiTerminal.attribute(E_NAME_TYPES).getText();
         List<?> colors = ansiTerminal.elements();

         Map<String, Color> colorMap = new HashMap<String, Color>();
         for (int k = 0; k < colors.size(); k++)
         {
            if (!(colors.get(k) instanceof DefaultElement))
            {
               continue;
            }

            DefaultElement colorElement = (DefaultElement) colors.get(k);
            String name = colorElement.attributeValue(E_NAME_NAME);
            Color color = parseColor(colorElement);
            colorMap.put(name, color);
         }

         for (String type : types.split(","))
         {
            output.put(type.trim(), colorMap);
         }

      }

      return output;
   }

   private Color parseColor(DefaultElement colorElement)
   {
      String foreground = colorElement.attributeValue(E_NAME_FOREGROUND).trim().toUpperCase();
      String background = colorElement.attributeValue(E_NAME_BACKGROUND).trim().toUpperCase();
      String attributes = colorElement.attributeValue(E_NAME_ATTRIBUTES);

      int attribute = attributes == null ? 0 : parseAttributes(attributes.split(","));
      ColorPalette parsedForeground = ColorPalette.valueOf(foreground);
      ColorPalette parsedBackround = ColorPalette.valueOf(background);

      return new Color(attribute, Color.toColorValue(parsedForeground, parsedBackround));
   }

   private int parseAttributes(String[] attributes)
   {
      int attribute = 0;

      for (String attributeName : attributes)
      {
         attribute |= ColorAttribute.valueOf(attributeName.trim().toUpperCase()).getValue();
      }

      return attribute;
   }

   private Document parseDocument()
   {
      try
      {
         EntityResolver resolver = new EntityResolver()
         {
            public InputSource resolveEntity(String publicId, String systemId)
            {
               if (systemId.endsWith(COLOR_MAP_DTD_ID))
               {
                  InputStream in = getClass().getResourceAsStream(COLOR_MAP_DTD_RESOURCE);
                  if (in == null)
                  {
                     LOG.warn("Color map DTD resource not found at " + COLOR_MAP_DTD_RESOURCE);
                     return null;
                  }
                  return new InputSource(in);
               }
               return null;
            }
         };

         InputStream inputStream = ClassLoader.class.getResourceAsStream(COLOR_MAP_XML_RESOURCE);
         if (inputStream == null)
         {
            LOG.warn("Color map XML resource not found at " + COLOR_MAP_XML_RESOURCE);
            return null;
         }
         
         SAXReader reader = new SAXReader(true);
         reader.setEntityResolver(resolver);
         Document document = reader.read(inputStream);
         return document;
      } 
      catch (DocumentException e)
      {
         LOG.error("Failed to load the color map XML", e);
         return null;
      }
   }
}
