/*
** Module   : ChuiScreenDriver.java
** Abstract : A base class for ChUI screen driver.
**
** Copyright (c) 2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- -----------------------Description------------------------
** 001 MAG 20131114 Created initial version.
** 002 EVL 20131212 Added method to check if the driver is in batch mode.
*/

package com.goldencode.p2j.ui.client.driver;

import java.util.*;

import com.goldencode.p2j.cfg.*;
import com.goldencode.p2j.main.*;
import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.chui.driver.*;
import com.goldencode.p2j.ui.client.chui.driver.console.KeyProcessor;
import com.goldencode.p2j.ui.client.chui.driver.swing.*;

/**
 * Base class for  ChUI driver.
 */
public class ChuiScreenDriver
implements ScreenDriver
{
   /** Direct to screen output primitives. */
   private OutputPrimitives direct;

   /** Terminal simulator implemented in Swing. */
   private ChuiSimulator sim = null;
   
   /** Bogus terminal type name. */
   private String termType = "vt100";

   /** This driver-specific widget factory */
   private WidgetFactory factory = new ChuiWidgetFactory();
   
   /**
    * Set <code>ChuiSimulator</code> instance
    * 
    * @param   sim
    *          Instance of <code>ChuiSimulator</code> simulator.
    */
   public void setChuiSimulator(ChuiSimulator sim)
   {
      this.sim = sim;      
   }

   /**
    * Set <code>OutputPrimitives</code> instance for drawing
    * 
    * @param   direct
    *          Primitives instance
    */
   public void setPrimitives(OutputPrimitives direct)
   {
      this.direct = direct;
   }
   
   /**
    * Provide an audible feedback to user by generating a beep signal.
    * <p>
    * There is no specific requirements to this method and actual implementation
    * may do nothing if there is no way to perform audible notification to the 
    * user.
    */
   @Override
   public void beep()
   {
      // Override and provide specific implementation.
   }

   /**
    * Clear screen.
    * <p>
    * This method should clear the screen and perform necessary flushing of the 
    * system-specific buffers if necessary. Users of this method assume that 
    * actual screen appearance as it is visible to the user is properly changed
    * upon exit from this method.
    */
   @Override
   public void clear()
   {
      sim.clear();
   }

   /**
    * Get driver-specific direct output primitives implementation.
    *
    * @return  driver-specific instance of direct screen output primitives.
    */
   @Override
   public OutputPrimitives getPrimitives()
   {
      return direct;
   }

   /**
    * Check if driver supports color output.
    * <p>
    * This method may be used to check if color output is supported by 
    * particular driver implementation.
    * <p>
    * TODO: at present there is no references to this method, can it be dropped
    * completely?
    * 
    * @return  <code>true</code> if colors are supported.
    */
   @Override
   public boolean hasColors()
   {
      return true;
   }

   /**
    * Checks if the driver is running in batch mode.
    * 
    * @return  <code>true</code> for batch mode, <code>false</code> otherwise.
    * 
    */
   public boolean inBatchMode()
   {
      return false;
   }
   
   /**
    * Read single keystroke from keyboard. Returned code should be either proper
    * Progress 4GL/P2J key code or -1 if environment-specific input cannot be 
    * properly translated.
    * <p>
    * The list of codes and detailed description is provided in {@link Keyboard}
    * class. The {@link KeyProcessor} can serve as an example of the key 
    * translation implementation.
    * 
    * @return  key code or -1 if no keys were pressed since last call.
    */
   @Override
   public int readKey()
   {
      return sim.readKey();
   }

   /**
    * Reset current screen mode.
    * <p>
    * Upon exit from this method P2J assumes that screen is in correct state
    * and can be used for interaction with the user.
    */   @Override
   public void resetMode()
   {
      // NOP for now, what should this do?
   }

    /**
     * Resume normal driver operation. If implementation is terminal-based,
     * they terminal should be acquired again and driver should be ready to 
     * perform regular screen I/O operations.
     */
   @Override
   public void resume()
   {
      // NOP for now, what should this do?
   }

   /**
    * Set cursor status (ON/OFF).
    * <p>
    * The implementation of this method should show/hide actual screen 
    * cursor visible to user. Remaining code assumes that actual cursor status
    * visible to user is properly changed upon exit from this method.
    * 
    * @param   on
    *          <code>true</code> - show cursor, <code>false</code> - hide cursor.
    * @return  <code>true</code> if operation was successful.
    * 
    * @throws  IllegalStateException
    */
   @Override
   public boolean setCursorStatus(boolean on)
   throws IllegalStateException
   {
      return sim.setCursorStatus(on);
   }

   /**
    * Change terminal type to specified one. The terminal-based implementation
    * should try to change terminal type to specified one. It is assumed that 
    * passed terminal string is specific to particular driver implementation 
    * and recognized by it. 
    * 
    * @param   name
    */
   @Override
   public void setTerminalType(String name)
   {
      termType = name;
   }

   /**
    * Suspend all screen I/O operations. If implementation is terminal-based, 
    * then terminal should be released to allow other application executed 
    * from P2J access terminal.
    */
   @Override
   public void suspend()
   {
      // NOP for now, what should this do?
   }

   /**
    * Obtain system/implementation specific terminal type string.
    * <p>
    * At present this is used to implement 4GL TERMINAL function and do some 
    * terminal-specific workarounds. At present workarounds exist for "vt320" 
    * terminal type. If implementation requires no special workarounds, then 
    * it should avoid returning "vt320" as terminal type string.  
    * 
    * @return  terminal type string.
    */
   @Override
   public String terminalType()
   {
      return termType;
   }

   /**
    * Driver initialization.
    */
   @Override
   public void init()
   {
      direct = new ChuiPrimitives(sim);
   }
   
   /**
    * Get driver-specific widget factory.
    * 
    * @return  reference to widget factory.
    */
   @Override
   public WidgetFactory getFactory()
   {
      return factory ;
   }
   
   /**
    * Get the default WINDOW-SYSTEM used by this driver. See {@link ClientParameters} for the list
    * of possible values.
    * 
    * @return   See above.
    */
   @Override
   public String getWindowSystem()
   {
      return ClientParameters.WINDOW_SYSTEM_TTY;
   }
   
   /**
    * Read configuration values from the bootstrap configuration and convert
    * these into an options instance.
    * <p>
    * The following parameters are supported:
    * <p>
    * <pre>
    * Category:Group:Key       Type
    * -----------------------  ---------------------------
    * client:chui:rows         integer
    * client:chui:columns      integer
    * client:chui:background   text which can be converted to a Color by Color.decode()
    * client:chui:foreground   text which can be converted to a Color by Color.decode()
    * client:chui:selection    text which can be converted to a Color by Color.decode()
    * client:chui:fontname     text containing a monospaced (fixed width) font family name
    * client:chui:fontsize     integer
    * </pre>
    *
    * @param    cfg
    *           The bootstrap configuration.
    *
    * @return   A new options instance with any defaults overridden as specified
    *           in the bootstrap configuration.
    */
   public static TerminalOptions convertToOptions(BootstrapConfig cfg)
   {
      TerminalOptions options = new TerminalOptions();
      
      String next;
      
      if ((next = cfg.getString("client", "chui", "rows", null)) != null)
      {
         options.rows = parseInt(next, options.rows);
      }
      
      if ((next = cfg.getString("client", "chui", "columns", null)) != null)
      {
         options.columns = parseInt(next, options.columns);
      }
      
      if ((next = cfg.getString("client", "chui", "background", null)) != null)
      {
         options.background = parseColor(next, options.background);
      }
      
      if ((next = cfg.getString("client", "chui", "foreground", null)) != null)
      {
         options.foreground = parseColor(next, options.foreground);
      }
      
      if ((next = cfg.getString("client", "chui", "selection", null)) != null)
      {
         options.selection = parseColor(next, options.selection);
      }
      
      if ((next = cfg.getString("client", "chui", "fontname", null)) != null)
      {
         options.fontname = next;
      }
      
      if ((next = cfg.getString("client", "chui", "fontsize", null)) != null)
      {
         options.fontsize = parseInt(next, options.fontsize);
      }
      
      return options;
   }
   
   /**
    * Attempt to convert the given string into an integer using
    * <code>Integer.parseInt</code>.
    *
    * @param    spec
    *           String encoded number specification.
    * @param    def
    *           Default value to return if the spec cannot be parsed properly.
    *
    * @return   The parsed number or the default if there is any parsing
    *           problem.
    */
   private static int parseInt(String spec, int def)
   {
      int val = def;
      
      try
      {
         val = Integer.parseInt(spec);
      }
      
      catch (NumberFormatException nfe)
      {
         // ignore
      }
      
      return val;
   }
   
   /**
    * Attempt to convert the given string into a color using
    * <code>Color.decode</code>.
    *
    * @param    spec
    *           String encoded color specification.
    * @param    def
    *           Default value to return if the spec cannot be parsed properly.
    *
    * @return   The parsed color or the default if the specification was
    *           invalid.
    */
   private static java.awt.Color parseColor(String spec, java.awt.Color def)
   {
      java.awt.Color color = def;
      
      try
      {
         color = java.awt.Color.decode(spec);
      }
      
      catch (NumberFormatException nfe)
      {
         // ignore
      }
      
      return color;
   }   
}
