/*
** Module   : terminal_linux.c
** Abstract : Linux implementation for character mode-based console functions.
**
** Copyright (c) 2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --------------------------------Description-----------------------------------
** 001 EVL 20130723 Created initial version to implement text console terminal JNI.
** 002 EVL 20130729 Isolating Linux specific code here. 
*/

#include "terminal.h"
#include <ncurses.h>
#include <term.h>
#include <termios.h>

// pointer to the structure of the interactive terminal
static SCREEN* terminal = NULL;

static int hascolors = 0;	      // set when we initialize

// The array of the character attributes to use in text drawing
jint attribute_result[ATTRIBUTE_MAX] =
   {A_NORMAL, A_STANDOUT, A_UNDERLINE, A_REVERSE, A_BLINK, A_DIM, A_BOLD, A_ALTCHARSET, A_INVIS};

// The array of the special symbols to draw the horizontal and vertical lines.
jint acs_result[ACS_MAX];

// terminal size at the moment of initialization of the library 
extern int base_rows;
extern int base_cols;

/**
 * Linux specific implementation of the moving cursor to the specified position in console.
 * The next character on the console will be placed in this location.
 *
 * @param   yRow     new Y(row) coordinate of the next char to draw.
 *
 * @param   xColumn  new X(column) coordinate of the next char to draw.
 */
void moveNative(int yRow, int xColumn)
{
   move(yRow, xColumn);
}

/**
 * Linux specific implementation of the character drawing to the current position in console.
 *
 * @param   chrToDraw     character to draw in the current cell.
 */
void addchNative(int chrToDraw)
{
   addch(chrToDraw);
}

/**
 * Returns the color attribute associated to the combination
 * of the supplied foreground and background colors.
 *
 * The function accepts the color range of <0..2*COLORS) where
 * the range of <0..COLORS) denotes the standard (usually 8) colors,
 * and the range <COLORS..2*COLORS) their highlighted variants.
 *
 * @param   f  foreground color
 *
 * @param   b  background color
 *
 * @return  attribute suitable for use in the ncurses functions
 *          accepting attributes (like attron)
 */
int getColorAttribute(int f, int b)
{
   if (!hascolors)
   {
      return 0;
   }

   // calculate the color pair (index) from the supplied standard
   // foreground and background colors.
   // A_BOLD is used to get the highlighted foreground,
   // A_BLINK is used to get the highlighted background,
   // note that A_BLINK will likely not produce all the color
   // variants for the background.
   return COLOR_PAIR(((b % COLORS) * COLORS) + (f % COLORS) + 1)
         | ((f < COLORS) ? 0 : A_BOLD) | ((b < COLORS) ? 0 : A_BLINK);
}

/**
 * Linux specific implementation of the calculating the color according to the given color pair
 * index.  The new attribute returned will be or-ed with the resulting color.
 *
 * @param   attrInitial     initial value of the attribute to work with.
 *
 * @param   colorPair       the color pair index to be mixed with initial attribute value.
 *
 * @return   The mixed attribute value as result of initial attribute and calculated color.
 */
int colorPairNative(int attrInitial, int colorPair)
{
   if (!hascolors)
   {
      // ignore the color if no color support
      return attrInitial;
   }

   int colorAttribute = 0;
   if (colorPair < 0)
   {
      // use the default color when skip-flag detected
      colorAttribute = COLOR_PAIR(0);
   } else
   {
      int foreground = colorPair >> 16;
      int background = colorPair & 0x0000FFFF;
      colorAttribute = getColorAttribute(foreground, background);
   }

   return attrInitial | colorAttribute;
}

/**
 * Beeps the tone on the terminal.
 */
void beepNative()
{
   beep();
}

/**
 * Clears the current console screen.
 */
void clearNative()
{
   clear();
}

/**
 * Returns the safe termname without null string on return.
 *
 * @return   The terminal type or "" on any failure.
 */
char* safe_termname()
{
   char* term = termname();
   return (term == NULL) ? "" : term;
}

/**
 * Get X coordinate of the cursor.
 *
 * @return   The integer value of the X coordinate.
 */
jint getx()
{
   int x, y;

   getyx(stdscr, y, x);
   return x;
}

/**
 * Get Y coordinate of the cursor.
 *
 * @param   env     pointer to a structure that provides access to JNI environment.
 *
 * @param   jo      reference to the Java object who called this function.
 *
 * @return   The integer value of the Y coordinate.
 */
jint gety()
{
   int x, y;

   getyx(stdscr, y, x);
   return y;
}

/**
 * Check if the terminal is able to display colors or not.
 *
 * @return   <code>true</code> if display is color capable, <code>false</code> otherwise.
 */
jboolean hasColors()
{
   return has_colors();
}

/**
 * Initializes terminal color support. The function initializes
 * COLORS*COLORS number of foreground/background color-pairs.
 * This makes all the standard colors available for use.
 */
void initColors()
{
   start_color();

   hascolors = has_colors();
   if (!hascolors)
   {
      return;
   }

   int f, b;

   // create foreground/background pairs of the standard number of COLORS
   for (f = 0; f < COLORS; ++f)
   {
      for (b = 0; b < COLORS; b++)
      {
         // the pair at 0 cannot be changed
         int res = init_pair((b * COLORS) + f + 1, f, b);
      }
   }
}

/**
 * This method performs console initialization on the native side.
 *
 * @param   env     pointer to a structure that provides access to JNI environment.
 *
 */
void initConsole(JNIEnv *env, char * termname)
{
   // use this instead of initscr because there is no way to query the
   // current terminal and when you use multiple terms, newterm overwrites
   // the current terminal... so you can never get it if you don't use
   // newterm directly
   terminal = newterm(termname, stdout, stdin);

   if (terminal == NULL)
   {
      char* spec   = "Interactive terminal of type %s cannot be initialized.";
      char* errmsg = malloc(strlen(spec) + strlen(termname) + 1); 
      sprintf(errmsg, spec, termname);
      throwException(env, "java/lang/IllegalStateException", errmsg);
      return;
   }
    
   // ACS_* are not constans and can not be included into static array declarations and should
   // be calculated after new terminal has been created.
   acs_result[0]  = ACS_ULCORNER;
	acs_result[1]  = ACS_LLCORNER;
	acs_result[2]  = ACS_URCORNER;
	acs_result[3]  = ACS_LRCORNER;
	acs_result[4]  = ACS_LTEE;
	acs_result[5]  = ACS_RTEE;
	acs_result[6]  = ACS_BTEE;
	acs_result[7]  = ACS_TTEE;
	acs_result[8]  = ACS_HLINE;
	acs_result[9]  = ACS_VLINE;
	acs_result[10] = ACS_PLUS;
	acs_result[11] = ACS_S1;
	acs_result[12] = ACS_S9;
	acs_result[13] = ACS_DIAMOND;
	acs_result[14] = ' '; // ACS_CKBOARD
	acs_result[15] = ACS_DEGREE;
	acs_result[16] = ACS_PLMINUS;
	acs_result[17] = ACS_BULLET;
	acs_result[18] = ' ';
	acs_result[19] = ' ';
	acs_result[20] = ' ';
	acs_result[21] = ' ';
	acs_result[22] = '>';
    
   keypad(stdscr, TRUE);      // enable keyboard mapping
   timeout(500);              // wait up to 500msec for input
   noecho();                  // don't echo input
   raw();
   nonl();
   auto_getch_refresh(FALSE); // disable sync during getch()
   def_prog_mode();           // save our state (after our mode changes)

   initColors();

   getmaxyx(stdscr, base_rows, base_cols);

   atexit((void (*)(void))endwin);
}

/**
 * This method reads the key from keyboard by the special reading thread.
 *
 * @return   The integer code of the key that was read or -1 if there is no character to read.
 */
jint readKey()
{
   return (jint) getch();
}

/**
 * Reset terminal settings.  May be called only if suspend() was previously called. 
 */
void resetMode()
{
   reset_prog_mode();
}

/**
 * Re-activates the console mode after ending of the previous suspension state.
 */
void resume()
{
   resetMode();                /* Return to the previous tty mode*/
                               /* stored by def_prog_mode()      */
   doupdate();                 /* Do refresh() to restore the    */
                               /* Screen contents                */
}

/**
 * Set absolute cursor position
 *
 * @param   x       new X coordinate of the cursor.
 *
 * @param   y       new Y coordinate of the cursor.
 *
 * @return   <code>true</code> if cursor positioning succeeded.
 */
jboolean setCursorNative(jint x, jint y)
{
   int localX, localY;
   move(y, x);

   // Set current cursor position
   getyx(stdscr, localY, localX);
    
   return localY == y && localX == x;
}

/**
 * Set the cursor visibility on or off. This method only affects the visibility of the cursor.
 *
 * @param   env     pointer to a structure that provides access to JNI environment.
 *
 * @param   on      when <code>true</code> the cursor will become visible, when <code>false</code>
 *                  the cursor will become invisible.
 *
 * @return   The previous state of the cursor (<code>true</code> means visible).
 */
jboolean setCursorStatus(JNIEnv *env, jboolean on)
{
   int rc = ERR;
   jboolean ret;
   
   if (on)
   {
      rc = curs_set( 1 );
   }
   else
   {
      rc = curs_set( 0 );
   }
   
   if (rc == 1 || rc == 2)
   {
      ret = true;
   }
   else if (rc == 0)
   {
      ret = false;
   }
   else
   {
      char* spec;
      char* termname = getenv("TERM");
      if (rc == ERR)
      {
         return true;
      }
      else
      {
         spec = "Unexpected return from curs_set() for terminal of type %s (%d).";
      }
      char* errmsg = malloc(strlen(spec) + strlen(termname) + 32); 
      sprintf(errmsg, spec, termname, rc);
      throwException(env, "java/lang/IllegalStateException", errmsg);
   }
   
   return ret;
}

/**
 * Simple reset of terminal type.
 *
 * @param   env       pointer to a structure that provides access to JNI environment.
 *
 * @param   termname  The name of the terminal to be restarted.
 */
void restartTerminal(JNIEnv *env, char* termname)
{
   int   err  = 0;
   int   rc   = restartterm(termname, fileno(stdout), &err); 
   
   if (rc == ERR)
   {
      char* spec   = "Failure resetting terminal of type %s (error %i).";
      char* errmsg = malloc(strlen(spec) + strlen(termname) + 15); 
      sprintf(errmsg, spec, termname, err);
      throwException(env, "java/lang/IllegalStateException", errmsg);
   }
}

/**
 * Save the current state of the console mode and switch to TTY mode temporarily.
 */
void suspend()
{
   def_prog_mode();            /* Save the tty modes             */
   endwin();                   /* End curses mode temporarily    */
}

/**
 * Synchronizes the state of the console buffers with native terminal window. Meaning the kind of
 * refresh calls depending on the backend used.
 */
void syncNative()
{
   refresh();    
}
